package calendar;

import observerPattern.Observer;
import observerPattern.Subject;

/**
 * The class for actual calendars. The contents of the calendar is an
 * unstructured sequence of <code>CalendarEntry</code> objects. In addition,
 * each calendar always has a view/index of type <code>SortedCalendarView</code>
 * that provides more advanced access methods. A calendar will observe all of
 * its entries and update itself whenever one of the entries changes. On the
 * other hand, it is possible to register to the calendar as a listener and be
 * notified about added and removed entries (modification of an entry appears as
 * removal and addition of the entry).
 */
public class Calendar extends CalendarModificationServer
                      implements CalendarView, Observer {

	//@ private invariant entries != null;
	//@ private invariant completeView != null;
	
    private final CalendarEntrySeq entries = new CalendarEntrySeq ();
    private final SortedCalendarView completeView;
    
    public Calendar() {
        completeView = new SortedCalendarView ( this );
    }
    
    /**
     * Get the sequence of all entries of the calendar
     */
    public CalendarEntrySeq getEntries() {
        return entries;
    }

    /**
     * Add a new entry to the calendar. It is possible to add an entry to more
     * than one calendar, or to add an entry more than once to one calendar
     * (although this is not very meaningful).
     */
    /*@
      @ public normal_behavior
      @  requires entry != null;
      @*/
    public void add(CalendarEntry entry) {
        entry.registerObserver ( this );
        addHelp ( entry );
    }

    /**
     * Remove an entry (once) from the calendar.
     */
    /*@
      @ public normal_behavior
      @  requires entry != null;
      @*/
    public void remove(CalendarEntry entry) {
        entry.unregisterObserver ( this );
        removeHelp ( entry );
    }

    /**
     * Get a view of the whole calendar that has somewhat more advanced access
     * methods than the ones provided here.
     */
    public /*@ pure @*/ SortedCalendarView getCompleteView() {
        return completeView;
    }

    private void addHelp(CalendarEntry entry) {
        entries.insert ( entry );
        fireAddedEntry ( entry );
    }

    private void removeHelp(CalendarEntry entry) {
        entries.removeFirst ( entry );
        fireRemovedEntry ( entry );
    }

    // Signals from the <code>Observer</code> interface
    
    public void prepareForModification(Subject subj) {
        removeHelp ( (CalendarEntry)subj );
    }
    
    public void notify(Subject subj) {
        addHelp ( (CalendarEntry)subj );
    }

}
