package calendar;

/**
 * Abstract superclass for classes that provide the
 * <code>CalendarListener</code> service.
 */
abstract class CalendarModificationServer {

    private CalendarListener[] listeners = new CalendarListener[1];
    private int listenersNum = 0;

    //@ private invariant listeners != null;
    //@ private invariant 1 <= listeners.length;
    //@ private invariant 0 <= listenersNum;
    //@ private invariant listenersNum <= listeners.length;
    /*@ 
      @ private invariant (\forall int i; 0 <= i && i < listenersNum;
      @                    listeners[i] != null);
      @*/
    
    /*@
      @ public normal_behavior
      @  requires listener != null;
      @  ensures listenersNum == \old(listenersNum) + 1;
      @  ensures listeners[listenersNum - 1] == listener &&
      @          (\forall int i; 0 <= i && i < listenersNum - 1;
      @              listeners[i] == \old(listeners[i]));
      @  assignable listenersNum, listeners, listeners[*];
      @*/
    public void registerListener(CalendarListener listener) {
        ++listenersNum;
        if ( listenersNum > listeners.length ) {
            final CalendarListener[] oldAr = listeners;
            listeners = new CalendarListener[listeners.length * 2];
            System.arraycopy ( oldAr, 0, listeners, 0, oldAr.length );
        }
    
        listeners[listenersNum - 1] = listener;
    }

    public void unregisterListener(CalendarListener listener) {
        int i = 0;
        while ( i < listenersNum && listeners[i] != listener ) {
            ++i;
        }
    
        if ( i < listenersNum ) {
            --listenersNum;
            
            int j = i;
            while ( j < listenersNum ) {
                listeners[j] = listeners[j + 1];
                ++j;
            }
            
            listeners[listenersNum] = null;
        }
    }

    protected void fireAddedEntry(CalendarEntry entry) {
        int i = 0;
        while ( i != listenersNum ) {
            listeners[i].addedEntry ( entry );
            ++i;
        }
    }

    protected void fireRemovedEntry(CalendarEntry entry) {
        int i = 0;
        while ( i != listenersNum ) {
            listeners[i].removedEntry ( entry );
            ++i;
        }
    }

    public /*@ pure @*/ int getListenersNum() {
        return listenersNum;
    }

    public /*@ pure @*/ CalendarListener getListener(int i) {
        return listeners[i];
    }
}
