package calendar;

import intervals.Interval;
import intervals.IntervalSeq;
import intervals.IntervalTree;

/**
 * A concrete calendar view that implements a secondary index on the calendar
 * entries. The class provides an efficient access method
 * <code>getOverlappingEntries</code> that retrieves all entries that overlap
 * a given interval. The view will always show exactly the same calendar entries
 * as the original calendar.
 */
public class SortedCalendarView implements CalendarView, CalendarListener {
    
    //@ private invariant calendar != null;
    //@ private invariant entryTree != null;
    
    private final Calendar calendar;
    private final IntervalTree entryTree = new IntervalTree ();
    
    public SortedCalendarView(Calendar calendar) {
        this.calendar = calendar;
        
        final CalendarEntrySeq entries = calendar.getEntries ();
        int i = 0;
        while ( i < entries.size () ) {
            entryTree.insert ( entries.at ( i ) );
            ++i;
        }
        
        calendar.registerListener ( this );
    }

    /**
     * Get the sequence of all entries of the calendar
     */
    public CalendarEntrySeq getEntries() {
        return calendar.getEntries ();
    }

    /**
     * Get a sequence of all entries of the calendar that overlap (have a point
     * of time in common with) the interval <code>iv</code>
     */
    public CalendarEntrySeq getOverlappingEntries(Interval iv) {
        return toCalendarEntrySeq ( entryTree.getOverlappingIntervals ( iv ) );
    }

    public void printStatistics() {
        System.out.println ( "Tree size: " + entryTree.treeSize() );
        System.out.println ( "Tree depth: " + entryTree.depth() );
        System.out.println ( "Number of entries: " + entryTree.size() );
    }

    private CalendarEntrySeq toCalendarEntrySeq(IntervalSeq intervals) {
        final CalendarEntrySeq res = new CalendarEntrySeq ();
        int i = 0;
        while ( i < intervals.size () ) {
            res.insert ( (CalendarEntry)intervals.at ( i ) );
            ++i;
        }
        return res;
    }
    
    // Signals from <code>CalendarListener</code> interface

    public void addedEntry(CalendarEntry e) {
        entryTree.insert ( e );
    }

    public void removedEntry(CalendarEntry e) {
        entryTree.remove ( e );
    }
}
