package intervals;

/**
 * Interval tree data structure for managing a finite multi-set of intervals.
 * The structure allows to efficiently retrieve the subset of intervals
 * overlapping a given point or interval.
 * 
 * Thomas H. Cormen, Charles E. Leiserson, Ronald L. Rivest, and Clifford Stein.
 * Introduction to Algorithms, Second Edition. MIT Press and McGraw-Hill, 2001.
 * ISBN 0-262-03293-7. Section 14.3: Interval trees, pp.311–317.
 */
public class IntervalTree {
    
    /**
     * The root node of the tree. Can be <code>null</code> for an empty set
     */
    private IntervalTreeNode root;
    
    /**
     * Insert one further interval in the set (an interval can be added more
     * than once)
     */
    /*@
      @ public normal_behavior
      @  requires iv != null;
      @*/
    public void insert(Interval iv) {
        root = IntervalTreeNode.insert ( root, iv );
    }
    
    /**
     * Remove the interval <code>iv</code> once from this set (it could occur
     * more than once as element of the tree)
     * 
     * @return <code>true</code> if the interval <code>iv</code> was element
     *         of the set
     */
    /*@
      @ public normal_behavior
      @  requires iv != null;
      @*/    
    public boolean remove(Interval iv) {
        return IntervalTreeNode.remove ( root, iv );
    }
    
    //@ public invariant size() >= 0;
    //@ public invariant (\forall int i; 0 <= i && i < size(); at(i) != null);
    
    /**
     * @return the number of elements of the multi-set
     */
    public /*@ pure @*/ int size() {
        if ( root == null ) return 0;
        return root.size ();
    }
    
    /**
     * Retrieve the <code>i</code>th interval of the multi-set. This method is
     * not efficient and is mostly used for specification purposes.
     */
    Interval /*@ pure @*/ at(int i) {
        if ( root == null ) return null;
        return root.at ( i );
    }    

    /**
     * @return all elements of the multi-set as a sequence
     */
    public IntervalSeq getIntervals() {
        final IntervalSeq res = new IntervalSeq ();
        if ( root != null ) root.collectIntervals ( res );
        return res;
    }    
    
    /**
     * Retrieve the elements that overlap a certain point <code>p</code>.
     * This method is efficient and has (provided that the tree is balanced,
     * which is not ensured by the present implementation) a complexity of O(m +
     * log n), where n is the number of elements of the set and m is the number
     * of overlapping intervals.
     * 
     * @return all elements of the multi-set that overlap the point
     *         <code>p</code>.
     */
    public IntervalSeq getOverlappingIntervals(int p) {
        return getOverlappingIntervals ( new SimpleInterval ( p, p ) );
    }

    /**
     * Retrieve the elements that overlap a certain interval <code>iv</code>.
     * This method is efficient and has (provided that the tree is balanced,
     * which is not ensured by the present implementation) a complexity of O(m +
     * log n), where n is the number of elements of the set and m is the number
     * of overlapping intervals.
     * 
     * @return all elements of the multi-set that overlap the interval
     *         <code>iv</code>.
     */
    public IntervalSeq getOverlappingIntervals(Interval iv) {
        final IntervalSeq res = new IntervalSeq ();
        if ( root != null ) root.collectOverlappingIntervals ( res, iv );
        return res;
    }

    /**
     * @return the number of nodes of the tree (only interesting for statistics,
     *         this has no meaning for the functionality of the tree)
     */
    public /*@ pure @*/ int treeSize() {
        if ( root == null ) return 0;
        return root.treeSize();
    }
    
    /**
     * @return the depth of the tree (only interesting for statistics, this has
     *         no meaning for the functionality of the tree)
     */
    public /*@ pure @*/ int depth() {
        if ( root == null ) return 0;
        return root.depth();
    }    
}
