package intervals;

/**
 * Class representing nodes of interval trees, see <code>IntervalTree</code>.
 * Each such node can have up to two children (binary tree), and stores two
 * lists of intervals that are sorted by the start and the end point (resp.).
 * The intervals stored in a node itself are exactly those intervals of the
 * subtree that overlap the point <code>cutPoint</code>
 */
class IntervalTreeNode {
    
    /**
     * The point that intervals have to overlap in order to be stored in this
     * node
     */
    private final int cutPoint;
    
    /**
     * The two children of this node
     */
    private IntervalTreeNode left;
    private IntervalTreeNode right;
    
    //@ public invariant sortedByStart != null;
    //@ public invariant sortedByEnd != null;
    //@ public invariant sortedByStart.size() == sortedByEnd.size();
    
    /**
     * Intervals overlapping <code>cutPoint</code>
     */    
    private final SortedByStartIntervalSeq sortedByStart
                        = new SortedByStartIntervalSeq ();
    private final SortedByEndIntervalSeq sortedByEnd
                        = new SortedByEndIntervalSeq ();
    
    IntervalTreeNode(int cutPoint) {
        this.cutPoint = cutPoint;
    }
    
    //@ public invariant treeSize() > 0;
    //@ public invariant size() >= 0;
    //@ public invariant (\forall int i; 0 <= i && i < size(); at(i) != null);

    /**
     * Number of nodes of the tree (including the root)
     */
    /*@
      @ public normal_behavior
      @  ensures \result == treeSize();
      @*/
    int /*@ pure @*/ treeSize() {
        return 1 + leftTreeSize () + rightTreeSize ();
    }

    /*@
      @ public normal_behavior
      @  ensures \result == leftTreeSize();
      @*/
    int /*@ pure @*/ leftTreeSize() {
        if ( left == null ) return 0;
        return left.treeSize ();
    }

    /*@
      @ public normal_behavior
      @  ensures \result == rightTreeSize();
      @*/
    int /*@ pure @*/ rightTreeSize() {
        if ( right == null ) return 0;
        return right.treeSize ();
    }

    /*@
      @ public normal_behavior
      @  ensures \result == size();
      @*/
    int /*@ pure @*/ size() {
        return sortedByStart.size () + leftSize() + rightSize();
    }

    /*@
      @ public normal_behavior
      @  ensures \result == leftSize();
      @*/
    int /*@ pure @*/ leftSize() {
        if ( left == null ) return 0;
        return left.size ();
    }
    
    /*@
      @ public normal_behavior
      @  ensures \result == rightSize();
      @*/
    int /*@ pure @*/ rightSize() {
        if ( right == null ) return 0;
        return right.size ();
    }
  
    int /*@ pure @*/ depth() {
        int res = 1;
        if ( left != null ) res = 1 + left.depth ();
        if ( right != null ) {
            final int rightDepth = 1 + right.depth ();
            if ( rightDepth > res ) return rightDepth;
        }
        return res;
    }
    
    /**
     * Function for enumerating the elements of this subtree. This method is
     * not efficient and is mostly used for specification purposes.
     */
    /*@
      @ public normal_behavior
      @  requires 0 <= i && i < sortedByStart.size();
      @  ensures \result == sortedByStart.at(i);
      @
      @ also
      @ public normal_behavior
      @  requires sortedByStart.size() <= i
      @           && i < sortedByStart.size() + leftSize();
      @  ensures \result == left.at ( i - sortedByStart.size() );
      @
      @ also
      @ public normal_behavior
      @  requires sortedByStart.size() + leftSize() <= i
      @           && i < sortedByStart.size() + leftSize() + rightSize();
      @  ensures \result == right.at ( i - sortedByStart.size() - leftSize() );
      @
      @ also
      @ public normal_behavior
      @  ensures \result == at(i);
      @*/
    Interval /*@ pure @*/ at(int i) {
        if ( i < 0 || i >= size () ) return null;
        IntervalTreeNode n = this;
        /*@
          @ loop_invariant n != null && i >= 0 && i < n.size();
          @ decreases n.treeSize();
          @ modifies n, i;
          @*/
        while ( true ) {
            if ( i < n.sortedByStart.size () ) return n.sortedByStart.at ( i );
            i -= n.sortedByStart.size ();
            if ( i < n.leftSize () ) {
                n = n.left;
            } else {
                i -= n.leftSize ();
                n = n.right;
            }
        }
    }

    /*@
      @ public invariant (\forall int i; 0 <= i && i < leftSize ();
      @                                  left.at ( i ).getEnd () < cutPoint);
      @ public invariant (\forall int i; 0 <= i && i < rightSize ();
      @                                  right.at ( i ).getStart () > cutPoint);
      @*/
    
    /**
     * Insert one further interval in the subtree (an interval can be added more
     * than once)
     * 
     * Missing here: balancing the resulting interval tree
     */
    /*@
      @ public normal_behavior
      @  requires iv != null;
      @  ensures size() == \old(size()) + 1;
      @  ensures (\exists int i; 0 <= i && i < size();
      @                          at(i) == iv
      @                          && (\forall int j; 0 <= j && j < i;
      @                                at(j) == \old(at(j)))
      @                          && (\forall int k; i < k && k < size();
      @                                at(k) == \old(at(k-1))));
      @  assignable \everything;
      @*/
    void insert(Interval iv) {
        IntervalTreeNode n = this;

        int c = n.compare ( iv );
        while ( c != 0 ) {
            if ( c < 0 ) {
                if ( n.left == null ) {
                    n.left = new IntervalTreeNode ( center ( iv ) );
                    n = n.left;                    
                    c = 0;
                } else {
                    n = n.left;
                    c = n.compare ( iv );
                }
            } else {
                if ( n.right == null ) {
                    n.right = new IntervalTreeNode ( center ( iv ) );
                    n = n.right;                    
                    c = 0;
                } else {
                    n = n.right;
                    c = n.compare ( iv );
                }
            }
        }
                
        n.sortedByStart.insert ( iv );
        n.sortedByEnd.insert ( iv );
    }

    static IntervalTreeNode insert(IntervalTreeNode node, Interval iv) {
        if ( node == null ) node = new IntervalTreeNode ( center ( iv ) );
        node.insert ( iv );
        return node;
    }

    /**
     * Compute the (bary)center of an interval.
     */
    /*@
      @ private normal_behavior
      @  requires iv != null;
      @  ensures iv.getStart() <= \result && \result <= iv.getEnd();
      @*/
    private static /*@ pure @*/ int center(Interval iv) {
        if ( iv.getStart () <= 0 && iv.getEnd () >= 0 ) {
            return ( iv.getStart () + iv.getEnd () ) / 2;
        } else {
            return ( iv.getStart () - iv.getEnd () ) / 2 + iv.getEnd ();
        }
    }
    
    /**
     * Remove the interval <code>iv</code> once from this subtree (it could
     * occur more than once as element of the subtree).
     * 
     * Missing here: there are no checks whether any nodes of the tree become
     * empty and can be removed.
     *
     * Missing here: balancing the resulting interval tree
     * 
     * @return <code>true</code> if the interval <code>iv</code> was element
     *         of the subtree
     */
    /*@
      @ public normal_behavior
      @  requires iv != null;
      @*/    
    boolean remove(Interval iv) {
        IntervalTreeNode n = this;

        int c = n.compare ( iv );
        while ( c != 0 ) {
            if ( c < 0 ) {
                if ( n.left == null ) return false;
                n = n.left;
            } else {
                if ( n.right == null ) return false;
                n = n.right;
            }
            c = n.compare ( iv );
        }

        n.sortedByStart.removeFirst ( iv );
        return n.sortedByEnd.removeFirst ( iv );
    }

    static boolean remove(IntervalTreeNode node, Interval iv) {
        if ( node == null ) return false;
        return node.remove ( iv );
    }

    /**
     * Collect all intervals that are stored in this subtree and insert them
     * in <code>seq</code>
     */
    void collectIntervals(IntervalSeq seq) {
        seq.insert ( sortedByStart );
        if ( left != null ) left.collectIntervals ( seq );
        if ( right != null ) right.collectIntervals ( seq );
    }    
    
    /**
     * Collect all intervals that are stored in this subtree and overlap the
     * interval <code>iv</code> and insert them in <code>seq</code>
     */
    void collectOverlappingIntervals(IntervalSeq seq, Interval iv) {
        if ( iv.getStart () < cutPoint ) {
            if ( left != null ) left.collectOverlappingIntervals ( seq, iv );
            sortedByStart.collectLeq ( seq, iv.getEnd () );
        } else {
            sortedByEnd.collectGeq ( seq, iv.getStart () );
        }

        if ( right != null && iv.getEnd () > cutPoint ) {
            right.collectOverlappingIntervals ( seq, iv );
        }
    }
    
    /**
     * @return a negative, zero, or positive number if <code>iv</code> is to
     *         the left, overlaps, or to the right of <code>cutPoint</code>
     */
    /*@
      @ private normal_behavior
      @  requires iv != null;
      @*/
    private /*@ pure @*/ int compare(Interval iv) {
        if ( iv.getStart () > cutPoint ) return 1;
        if ( iv.getEnd () < cutPoint ) return -1;
        return 0;
    }
    
}
