package observerPattern;

/**
 * Class representing observable objects (Observer Pattern). It is possible to
 * register to such objects and be notified about all modifications that occur.
 * This is a slightly modified variant of the observer pattern, in which there
 * are two signals: before an actual modification takes place, observers are
 * prepared by calling their <code>prepareForModification</code> method. After
 * the modification, the <code>notify</code> method of observers is called.
 * Subclasses of this class have to ensure that <code>notify</code> is always
 * raised exactly once after a <code>prepareForModification</code>.
 */
public class Subject {
    
    //@ private invariant observers != null;
    //@ private invariant 1 <= observers.length;
    //@ private invariant 0 <= observersNum;
    //@ private invariant observersNum <= observers.length;
    /*@ 
      @ private invariant (\forall int i; 0 <= i && i < observersNum;
      @                                   observers[i] != null);
      @*/
    
    private Observer[] observers = new Observer[1];
    private int observersNum = 0;
    
    public void registerObserver(Observer obs) {
        ++observersNum;
        if ( observersNum > observers.length ) {
            final Observer[] oldAr = observers;
            observers = new Observer[observers.length * 2];
            int i = 0;
            /*@
              @ loop_invariant 0 <= i && i <= oldAr.length &&
              @                (\forall int j; 0 <= j && j < i;
              @                                observers[j] == oldAr[j]);
              @ modifies i, observers[*];
              @ decreases oldAr.length - i;
              @*/
            while ( i < oldAr.length ) {
            	observers[i] = oldAr[i];
            	++i;
            }
        }

        observers[observersNum - 1] = obs;
    }
    
    /**
     * This is a HACK for verifying the unregisterObserver method; we work
     * around the missing key support for \old in loop invariants
     */
    
    //@ public model observerPattern.Observer[] oldAr;
    
    /*@
      @ public normal_behavior
      @   requires (\forall int i; 0 <= i && i < observersNum;
      @                            observers[i] != obs);
      @   assignable \nothing;
      @
      @ also
      @ public normal_behavior
      @   requires (\exists int i; 0 <= i && i < observersNum;
      @                            observers[i] == obs);
      @   requires oldAr != null && oldAr != observers
      @            && (\forall int i; 0 <= i && i < observersNum;
      @                               oldAr[i] == observers[i]);
      @   ensures observersNum == \old(observersNum) - 1;
      @   ensures (\exists int i; 0 <= i && i < \old(observersNum);
      @                           \old(observers[i]) == obs
      @                           && (\forall int j; 0 <= j && j < i;
      @                                 observers[j] == \old(observers[j]))
      @                           && (\forall int k; i < k && k < \old(observersNum);
      @                                 observers[k-1] == \old(observers[k])));
      @   assignable observers[*], observersNum;
      @*/
    public void unregisterObserver(Observer obs) {
        int i = 0;
        /*@
          @ loop_invariant 0 <= i && i <= observersNum
          @                && (\forall int k; 0 <= k && k < i;
          @                                   observers[k] != obs);
          @ decreases observersNum - i;
          @ modifies i;
          @*/
        while ( i < observersNum && observers[i] != obs ) {
            ++i;
        }

        if ( i < observersNum ) {
            --observersNum;

            int j = i;
            /*@
              @ loop_invariant i <= j && j <= observersNum
              @                && (\forall int k; (0 <= k && k < i) || ( j <= k && k <= observersNum);
              @                                   oldAr[k] == observers[k])
              @                && (\forall int k; i <= k && k < j;
              @                                   oldAr[k+1] == observers[k]);
              @ decreases observersNum - j;
              @ modifies j, observers[*];
              @*/
            while ( j < observersNum ) {
                observers[j] = observers[j + 1];
                ++j;
            }

            observers[observersNum] = null;
        }
    }
    
    protected void prepareObservers() {
        int i = 0;
        while ( i < observersNum ) {
            observers[i].prepareForModification ( this );
            ++i;
        }
    }

    protected void notifyObservers() {
        int i = 0;
        while ( i < observersNum ) {
            observers[i].notify ( this );
            ++i;
        }
    }
}
